{
Unit GenMaze.pas
----------------

Copyright (c) 2003

Licensed under the BSD/MIT License:

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}

unit GenMaze;

interface

uses Classes;

type
  TMazeAlg = (mzBacktrack, mzPrim, mzHunt);

  TWallExists = record
  	RightWall, BottomWall: Boolean;
  end;

  AAWallExists = Array of Array of TWallExists;

  TMazePointEx = record
  	X: Integer;
    Y: Integer;
    PrevX: Integer;
    PrevY: Integer;
    nSteps: Integer; { for Prim's Algorithm }
  end;

  PMazePointEx = ^TMazePointEx;

procedure GenerateMaze(var MazeWalls: AAWallExists; var Solution: TList;
			Algorithm: TMazeAlg);
procedure StubSolve(var MazeWalls: AAWallExists; var Solution: TList);

implementation

uses Types, SysUtils, Dialogs;

procedure BiasedPrim(var MazeWalls: AAWallExists; var Solution: TList); forward;
procedure PrimAlg(var MazeWalls: AAWallExists; var Solution: TList); forward;
procedure HuntAlg(var MazeWalls: AAWallExists; var Solution: TList;
			Algorithm: TMazeAlg); forward;

procedure GenerateMaze(var MazeWalls: AAWallExists; var Solution: TList;
			Algorithm: TMazeAlg);
begin
    case Algorithm of
    	mzBacktrack: HuntAlg(MazeWalls, Solution, mzBacktrack);
    	mzHunt: HuntAlg(MazeWalls, Solution, mzHunt);
		mzPrim: PrimAlg(MazeWalls, Solution);
    end;
end;

procedure BreakHWall(var MazeWalls: AAWallExists; x1, y1, x2, y2: Integer);
begin
	if (x1 >= 0) and (y1 >= 0) then
		MazeWalls[y1, x1].BottomWall := True;
end;

procedure BreakVWall(var MazeWalls: AAWallExists; x1, y1, x2, y2: Integer);
begin
	if (x1 >= 0) and (y1 >= 0) then
		MazeWalls[y1, x1].RightWall := True;
end;

{ See http://www.astrolog.org/labyrnth/algrithm.htm for a high-level
  discussion of the algorithms themselves }
{ All three algorithms are implemented roughly the same way now, with
  a major difference at only one step, that I will point out later.
  Starting at the upper-left corner, the algorithms cut out walls
  next to an existing, explored cell, with the direction selected
  at random.  In Prim's Algorithm, the cell itself is also selected
  completely randomly, subject only to the constraint that it is an
  explored cell.  In Recursive Backtracking, which is basically DFS
  (Depth First Search) applied to construction rather than search,
  it is always the last cell carved that is chosen.  And the Hunt
  algorithm is a combination of both--always select the last cell
  until you come to a dead end, then select randomly from the entire
  fringe (as opposed to DFS, which picks the last cell carved that
  is possible to brance out from). }
procedure PrimAlg(var MazeWalls: AAWallExists; var Solution: TList);
var
	i, j: Integer;
    MaxX, MaxY: Integer;
    Explored: Array of Array of Integer;
    PrevPoints: Array of Array of TPoint;
    nExplored, nStep, total: Integer;
    pt, newpt, CurPoint, PrevPoint: TPoint;
    PNewPoint: PMazePointEx;
begin
	{ Initialization of variables }
    SetLength(Explored, Length(MazeWalls));
    SetLength(PrevPoints, Length(MazeWalls));
	for j := Low(MazeWalls) to High(MazeWalls) do
    begin
    	SetLength(Explored[j], Length(MazeWalls[j]));
	    SetLength(PrevPoints[j], Length(MazeWalls[j]));
		for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        begin
            Explored[j, i] := -1;
            PrevPoints[j, i].X := -1;
            PrevPoints[j, i].Y := -1;
        end;
    end;
    total := Length(MazeWalls)*Length(MazeWalls[0]);
    MaxX := High(MazeWalls[0]);
    MaxY := High(MazeWalls);

	Explored[0, 0] := 0;
    nStep := 0;
    nExplored := 1;

	while nExplored < total do
    begin
    	while True do
        begin
        	pt.Y := Trunc(Random(Length(MazeWalls)));
            pt.X := Trunc(Random(Length(MazeWalls[pt.Y])));
            if Explored[pt.Y,pt.X] < 0 then
            begin
				newpt := pt;
            	if ((newpt.X < MaxX) and (Explored[newpt.Y, newpt.X + 1] >= 0)) then
                	pt.X := newpt.X + 1;
                if ((newpt.X > 0) and (Explored[newpt.Y, newpt.X - 1] >= 0)) then
                	if (pt.X = newpt.X) or (Random < 0.5) then
            			pt.X := newpt.X - 1;
            	if ((newpt.Y < MaxY) and (Explored[newpt.Y + 1, newpt.X] >= 0)) then
            		pt.Y := newpt.Y + 1;
            	if ((newpt.Y > 0) and (Explored[newpt.Y - 1, newpt.X] >= 0)) then
                	if (pt.Y = newpt.Y) or (Random < 0.5) then
            			pt.Y := newpt.Y - 1;

                { Not on fringe, so keep searching }
				if (newpt.X = pt.X) and (newpt.Y = pt.Y) then
	            	Continue;

				{ Break ties randomly, as above }
                if (newpt.X <> pt.X) and (newpt.Y <> pt.Y) then
                begin
                	if Random < 0.5 then
                    	pt.Y := newpt.Y
                    else
						pt.X := newpt.X;
                end;

              	if      newpt.X < pt.X then { going to the right }
               		BreakHWall(MazeWalls, newpt.X, newpt.Y, pt.X, pt.Y)
              	else if pt.X < newpt.X then { going to the left }
               		BreakHWall(MazeWalls, pt.X, pt.Y, newpt.X, newpt.Y)
              	else if newpt.Y < pt.Y then { going down }
               		BreakVWall(MazeWalls, newpt.X, newpt.Y, pt.X, pt.Y)
              	else if pt.Y < newpt.Y then { going up }
               		BreakVWall(MazeWalls, pt.X, pt.Y, newpt.X, newpt.Y);

                Explored[newpt.Y, newpt.X] := Explored[pt.Y, pt.X] + 1;
            	PrevPoints[newpt.Y, newpt.X].X := pt.X;
            	PrevPoints[newpt.Y, newpt.X].Y := pt.Y;
                Break;
            end;
        end; { while True }

    	nExplored := nExplored + 1;
    end; { while nExplred < Total }

    //StubSolve(MazeWalls, Solution);



    { From a freshly-generated maze, obtaining the solution is easy--simply
    	backtrack your way from the end to the beginning, using the previous
        step stored in each square. }
    CurPoint.X := MaxX;
    CurPoint.Y := MaxY;
    PrevPoint := PrevPoints[CurPoint.Y, CurPoint.X];
    //nStep := Explored[CurPoint.Y, CurPoint.X];
    //while (CurPoint.X > -1) and (CurPoint.Y > -1) do
    while True do
    begin
     	New(PNewPoint);
        PNewPoint^.X := CurPoint.X;
        PNewPoint^.Y := CurPoint.Y;
        PNewPoint^.PrevX := PrevPoint.X;
        PNewPoint^.PrevY := PrevPoint.Y;
        PNewPoint^.nSteps := Explored[CurPoint.Y, CurPoint.X];

        Solution.Add(PNewPoint);

        if (CurPoint.X = 0) and (CurPoint.Y = 0) then Break;

		CurPoint := PrevPoint;
		PrevPoint := PrevPoints[CurPoint.Y, CurPoint.X];

        //nStep := nStep - 1;
    end;

end;


procedure HuntAlg(var MazeWalls: AAWallExists; var Solution: TList;
			Algorithm: TMazeAlg);
var
	i, j, n: Integer;
    MaxX, MaxY: Integer;
    Explored: Array of Array of Integer;
    PrevPoints: Array of Array of TPoint;
    nExplored, nStep, total: Integer;
    pt, newpt, CurPoint, PrevPoint: TPoint;
    PNewPoint: PMazePointEx;
    Done: Boolean;
begin
	{ Initialization of variables }
    SetLength(Explored, Length(MazeWalls));
    SetLength(PrevPoints, Length(MazeWalls));
	for j := Low(MazeWalls) to High(MazeWalls) do
    begin
    	SetLength(Explored[j], Length(MazeWalls[j]));
	    SetLength(PrevPoints[j], Length(MazeWalls[j]));
		for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        begin
            Explored[j, i] := -1;
            PrevPoints[j, i].X := -1;
            PrevPoints[j, i].Y := -1;
        end;
    end;
    total := Length(MazeWalls)*Length(MazeWalls[0]);
    MaxX := High(MazeWalls[0]);
    MaxY := High(MazeWalls);

	pt.X := 0;
    pt.Y := 0;
	Explored[0, 0] := 0;
    nStep := 0;
    nExplored := 1;

	while nExplored < total do
    begin
    	{ Explore one more square out from where we are;
          if trapped, "hunt" for a new starting point on the fringe }

        if  (((pt.X > 0)    and (Explored[pt.Y, pt.X - 1] >= 0)) or (pt.X = 0)) and
        	(((pt.X < MaxX) and (Explored[pt.Y, pt.X + 1] >= 0)) or (pt.X = MaxX)) and
            (((pt.Y > 0)    and (Explored[pt.Y - 1, pt.X] >= 0)) or (pt.Y = 0)) and
            (((pt.Y < MaxY) and (Explored[pt.Y + 1, pt.X] >= 0)) or (pt.Y = MaxY)) then
        begin
        	if Algorithm = mzHunt then
            begin
            	{ Method #1: Randomly pick a square with a neighbor in the maze}
                repeat
                	j := Trunc(Random(Length(MazeWalls)));
                	i := Trunc(Random(Length(MazeWalls[j])));
                until (Explored[j,i] >= 0);
                pt.X := i;
                pt.Y := j;
                nStep := Explored[j,i];
                Continue;
            end
            else { Algorithm = mzBacktrack }
            begin
	        	{ Method #2: DFS-alike: go back to the previous square }
            	if ((pt.X < MaxX) and (Explored[pt.Y, pt.X + 1] = nStep - 1)) then
                	pt.X := pt.X + 1
                else if ((pt.X > 0) and (Explored[pt.Y, pt.X - 1] = nStep - 1)) then
            		pt.X := pt.X - 1
            	else if ((pt.Y < MaxY) and (Explored[pt.Y + 1, pt.X] = nStep - 1)) then
            		pt.Y := pt.Y + 1
            	else if ((pt.Y > 0) and (Explored[pt.Y - 1, pt.X] = nStep - 1)) then
        	    	pt.Y := pt.Y - 1
    	        else
	            	Break;

            	nStep := nStep - 1;
            	if nStep < 0 then
            	begin
            		ShowMessage('Error!!!!!!');
            		Break;
            	end;
            end; { algorithm choice }

            Continue;
        end;

        nStep := nStep + 1;

		Done := False;
		while not Done do
        begin
        	n := Trunc(Random(4));
            if n = 0 then
            	if ((pt.X < MaxX) and (not Explored[pt.Y, pt.X + 1] >= 0)) then
                begin
                    newpt := pt;
                    Inc(newpt.X);
					BreakHWall(MazeWalls, pt.X, pt.Y, newpt.X, newpt.Y);
                	Done := True;
                end;
            if n = 1 then
            	if ((pt.X > 0) and (not Explored[pt.Y, pt.X - 1] >= 0)) then
                begin
                    newpt := pt;
                    Dec(newpt.X);
					BreakHWall(MazeWalls, newpt.X, newpt.Y, pt.X, pt.Y);
                	Done := True;
                end;
            if n = 2 then
            	if ((pt.Y < MaxY) and (not Explored[pt.Y + 1, pt.X] >= 0)) then
                begin
                    newpt := pt;
                    Inc(newpt.Y);
					BreakVWall(MazeWalls, pt.X, pt.Y, newpt.X, newpt.Y);
                	Done := True;
                end;
            if n = 3 then
            	if ((pt.Y > 0) and (not Explored[pt.Y - 1, pt.X] >= 0)) then
                begin
                    newpt := pt;
                    Dec(newpt.Y);
					BreakVWall(MazeWalls, newpt.X, newpt.Y, pt.X, pt.Y);
                	Done := True;
                end;

        end;
		PrevPoints[newpt.Y, newpt.X] := pt;
        Explored[newpt.Y, newpt.X] := nStep;

		pt := newpt;
    	nExplored := nExplored + 1;
	end;

    { From a freshly-generated maze, obtaining the solution is easy--simply
    	backtrack your way from the end to the beginning, using the previous
        step stored in each square. }
    CurPoint.X := MaxX;
    CurPoint.Y := MaxY;
    PrevPoint := PrevPoints[CurPoint.Y, CurPoint.X];
    //nStep := Explored[CurPoint.Y, CurPoint.X];
    //while (CurPoint.X > -1) and (CurPoint.Y > -1) do
    while True do
    begin
     	New(PNewPoint);
        PNewPoint^.X := CurPoint.X;
        PNewPoint^.Y := CurPoint.Y;
        PNewPoint^.PrevX := PrevPoint.X;
        PNewPoint^.PrevY := PrevPoint.Y;
        PNewPoint^.nSteps := Explored[CurPoint.Y, CurPoint.X];

        Solution.Add(PNewPoint);

        if (CurPoint.X = 0) and (CurPoint.Y = 0) then Break;

		CurPoint := PrevPoint;
		PrevPoint := PrevPoints[CurPoint.Y, CurPoint.X];

        //nStep := nStep - 1;
    end;
end;

{ This is mainly a procedure to use when writing a new maze-generation
  algorithm; it sets the solution to a constant go-all-the-way-right-
  then-go-all-the-way-down regardless of the maze, so can be used when
  you don't care about getting the solution right.  Eventually, I might
  add a true, generalized DFS solving algorithm that can be used on
  any maze, but it is not needed as all algorithms so far generate the
  maze in such a way that the solution is trivial to generate from the
  data structures used in the creation of the maze. }
procedure StubSolve(var MazeWalls: AAWallExists; var Solution: TList);
var
	i, j, n: Integer;
    PointList: Array of Array of TMazePointEx;
    CurPoint: TMazePointEx;
	PNewPoint: PMazePointEx;
begin
    SetLength(PointList, Length(MazeWalls));
	for j := Low(MazeWalls) to High(MazeWalls) do
    begin
    	SetLength(PointList[j], Length(MazeWalls[j]));
		for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        begin
            with PointList[j, i] do
            begin
            	X := i;
            	Y := j;
                PrevX := 0;
                PrevY := 0;
                nSteps := 99999; { i.e.: Infinity }
            end;
        end;
    end;

    n := 0;
    for j := High(MazeWalls) downto 0 do
    begin
    	CurPoint.X := High(MazeWalls[0]);
        CurPoint.Y := j;
        CurPoint.nSteps := n;
        Inc(n);

        New(PNewPoint);
        PNewPoint^ := CurPoint;
        Solution.Add(PNewPoint);
    end;

    for i := High(MazeWalls[0]) downto 0 do
    begin
    	CurPoint.X := i;
        CurPoint.Y := 0;
        CurPoint.nSteps := n;
        Inc(n);

        New(PNewPoint);
        PNewPoint^ := CurPoint;
        Solution.Add(PNewPoint);
    end;

end;

{ An older, slower, and more maddening version of Prim's algorithm; it
  was the first one created, but had its flaws and was rewritten from
  scratch with the lessons of the Hunt algorithm in mind, especially
  that randomly finding a square is in practice much faster than using
  a TList to store the fringe, despite its theoretical disadvantage
  that it's not even garunteed to terminate!  So if you're one of the
  lucky one in a billion to have the current Prim's algorithm hang,
  simple kill the program and try again and it won't happen again
  until the sun goes supernova! }
function SelectSpace(List: TList): Integer; forward;

procedure BiasedPrim(var MazeWalls: AAWallExists; var Solution: TList);
var
	i, j, xx, yy, prevxx, prevyy: Integer;
    W, H: Integer;
    NewSpace: Integer;
    PointList: Array of Array of TMazePointEx;
    CurPoint: TMazePointEx;
	PNewPoint: PMazePointEx;
    Fringe: TList;  { Maintain the "fringe" of unexplored spaces }
    Explored: Array of Array of Boolean;
begin
	{ Algorithm is Randomized Search, roughly the algorithm
      explained on MazeWorks.  See:
      http://www.mazeworks.com/mazegen/mazetut/index.htm
      for more details.  Basically, keep tabs on the "fringes" of
      the explored space, and be careful to not step in the same
      space twice. }

	{ Initialization of variables }
    SetLength(Explored, Length(MazeWalls));
    SetLength(PointList, Length(MazeWalls));
	for j := Low(MazeWalls) to High(MazeWalls) do
    begin
    	SetLength(Explored[j], Length(MazeWalls[j]));
    	SetLength(PointList[j], Length(MazeWalls[j]));
		for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        begin
            Explored[j, i] := False;
            with PointList[j, i] do
            begin
            	X := i;
            	Y := j;
                PrevX := 0;
                PrevY := 0;
                nSteps := 99999; { i.e.: Infinity }
            end;
        end;
    end;

	{ The algorithm itself }
    with PointList[0, 0] do
    begin
    	X      :=  0;
    	Y      :=  0;
    	PrevX  := -1;
    	PrevY  :=  0;
        nSteps :=  0;
	end;
	Fringe := TList.Create;
	try
    	New(PNewPoint);
        PNewPoint^ := PointList[0, 0];
    	Fringe.Add(PNewPoint); { Start at upper-left corner }
		while Fringe.Count > 0 do
        begin
        	{ Pop a value out of the list }
        	NewSpace := SelectSpace(Fringe);
            CurPoint := PMazePointEx(Fringe.Items[NewSpace])^;
			Dispose(Fringe.Items[NewSpace]);
            Fringe.Delete(NewSpace);

            { Make sure it's not already explored; mark it as such }
			xx := CurPoint.X;
            yy := CurPoint.Y;
            prevxx := CurPoint.PrevX;
            prevyy := CurPoint.PrevY;
            if Explored[yy, xx] then
            	continue;
            PointList[yy, xx] := CurPoint;
            Explored[yy, xx] := True;
            {ShowMessage('(' + IntToStr(xx) + ', ' + IntToStr(yy) + ')' +
            	'           (' + IntToStr(CurPoint.PrevX) + ', ' +
                IntToStr(CurPoint.PrevY) + ')');}

            { Break down the wall between the previous square
              and the current one }
              if      xx < prevxx then { going to the right }
               	BreakHWall(MazeWalls, xx, yy, prevxx, Prevyy)
              else if prevxx < xx then { going to the left }
               	BreakHWall(MazeWalls, prevxx, prevyy, xx, yy)
              else if yy < prevyy then { going down }
               	BreakVWall(MazeWalls, xx, yy, prevxx, prevyy)
              else if prevyy < yy then { going up }
               	BreakVWall(MazeWalls, prevxx, prevyy, xx, yy);

            { Push all of its neighbors onto the fringe that haven't
              already been explored }
            if (yy in [Low(MazeWalls)..High(MazeWalls)]) and
            		((xx - 1) in [Low(MazeWalls[yy])..High(MazeWalls[yy])]) and
            		(not Explored[yy, xx - 1]) then
            begin
    			New(PNewPoint);
        		PNewPoint^ := PointList[yy, xx - 1];
                PNewPoint^.PrevX := xx;
                PNewPoint^.PrevY := yy;
                Inc(PNewPoint^.nSteps);
    			Fringe.Add(PNewPoint);
            end;
            if (yy in [Low(MazeWalls)..High(MazeWalls)]) and
            		((xx + 1) in [Low(MazeWalls[yy])..High(MazeWalls[yy])]) and
            		(not Explored[yy, xx + 1]) then
            begin
    			New(PNewPoint);
        		PNewPoint^ := PointList[yy, xx + 1];
                PNewPoint^.PrevX := xx;
                PNewPoint^.PrevY := yy;
                Inc(PNewPoint^.nSteps);
    			Fringe.Add(PNewPoint);
            end;
            if ((yy - 1) in [Low(MazeWalls)..High(MazeWalls)]) and
            		(not Explored[yy - 1, xx]) then
            begin
    			New(PNewPoint);
        		PNewPoint^ := PointList[yy - 1, xx];
                PNewPoint^.PrevX := xx;
                PNewPoint^.PrevY := yy;
                Inc(PNewPoint^.nSteps);
    			Fringe.Add(PNewPoint);
            end;
            if ((yy + 1) in [Low(MazeWalls)..High(MazeWalls)]) and
            		(not Explored[yy + 1, xx]) then
            begin
    			New(PNewPoint);
        		PNewPoint^ := PointList[yy + 1, xx];
                PNewPoint^.PrevX := xx;
                PNewPoint^.PrevY := yy;
                Inc(PNewPoint^.nSteps);
    			Fringe.Add(PNewPoint);
            end;
        end;

		{ From a freshly-generated maze, obtaining the solution is easy--simply
          backtrack your way from the end to the beginning, using the previous
      	  step stored in each square. }
		H := High(MazeWalls) + 1;
        W := High(MazeWalls[H - 1]) + 1;
        CurPoint := PointList[H - 1, W - 1];
        while (CurPoint.X > 0) or (CurPoint.Y > 0) do
        begin
        	New(PNewPoint);
        	PNewPoint^ := CurPoint;
        	Solution.Add(PNewPoint);

            CurPoint := PointList[CurPoint.PrevY, CurPoint.PrevX];
        end;
        New(PNewPoint);
        PNewPoint^ := CurPoint; { Which should be (0,0) now }
        Solution.Add(PNewPoint);
    finally
    	Fringe.Free;
    end;
end;

{ For use only with the BiasedPrim algorithm }
function SelectSpace(List: TList): Integer;
const
	BiasAmount = 3;
var
	a, b : Extended;
    Bias : Array of Extended;
    i, n, n2, nback, nforward, nn : Integer;
	PPoint, PLastPoint : PMazePointEx;
begin
	{ Result := Trunc(Random(List.Count)); }
    { For BFS: Result := 0; }
    { For DFS: Result := List.Count - 1; }
    if List.Count = 0 then
    begin
    	Result := 0;
        Exit;
    end;
    PLastPoint := List.Items[List.Count - 1];
    n := PLastPoint.nSteps;
    nn := 0;
    nback := 0;
    for i := 0 to List.Count - 1 do
    begin
		PPoint := List.Items[i];
        n2 := PPoint^.nSteps;
        if n = n2 then
        begin
        	Inc(nn);
            if (PPoint^.X < PPoint^.PrevX) or (PPoint^.Y < PPoint^.PrevY) then
            	Inc(nback);
        end;
    end;
    nforward := nn - nback;

    SetLength(Bias, List.Count);
    a := 1.0/(nforward + BiasAmount*nback);
    b := BiasAmount/(nforward + BiasAmount*nback);
    for i := 0 to List.Count - 1 do
    begin
		PPoint := List.Items[i];
        n2 := PPoint^.nSteps;
        if n = n2 then
        begin
        	if (PPoint^.X < PPoint^.PrevX) or (PPoint^.Y < PPoint^.PrevY) then
            begin
            	if i > 0 then
        			Bias[i] := b + Bias[i - 1]
                else
                	Bias[i] := b;
            end
            else
            begin
            	if i > 0 then
            		Bias[i] := a + Bias[i - 1]
                else
                	Bias[i] := a;
            end;
        end
        else
        begin
        	if i > 0 then
        		Bias[i] := Bias[i - 1]
            else
            	Bias[i] := 0;
        end;
    end;

    a := Random;
    {if a < 0.05 then
    	Result := List.Count - 1
    else if a < 0.10 then
		Result := Trunc(Random(List.Count))
    else
	repeat
		Result := Trunc(Random(List.Count));
		PPoint := List.Items[Result];
        n2 := PPoint^.nSteps;
    until n = n2;}
    Result := 0;
    for i := 0 to List.Count - 1 do
    begin
    	if a < Bias[i] then
        begin
        	Result := i;
            Break;
        end;
	end;
end;


end.
